<?php

namespace Drupal\Tests\media_pixabay\FunctionalJavascript;

use function GuzzleHttp\json_encode;
use Drupal\FunctionalJavascriptTests\WebDriverTestBase;
use Drupal\FunctionalJavascriptTests\WebDriverWebAssert;
use Drupal\Tests\media\Traits\MediaTypeCreationTrait;

/**
 * Tests the Pixabay widget adding and usage.
 *
 * @group Media Pixabay
 *
 * @package Drupal\Tests\media_pixabay\FunctionalJavascript
 */
class MediaPixabayWidgetTest extends WebDriverTestBase {

  use MediaTypeCreationTrait;

  /**
   * Modules to enable.
   *
   * @var array
   */
  protected static $modules = [
    'entity_browser',
    'media',
    'media_pixabay',
    'media_pixabay_test',
    'media_test_source',
  ];

  /**
   * {@inheritdoc}
   */
  protected $defaultTheme = 'stable';

  /**
   * Media type entity.
   *
   * @var \Drupal\media\MediaTypeInterface
   */
  protected $mediaType;

  /**
   * The test user.
   *
   * @var \Drupal\user\UserInterface
   */
  protected $user;

  /**
   * Media type storage.
   *
   * @var \Drupal\media\MediaStorage
   */
  protected $mediaTypeStorage;

  /**
   * Field config storage.
   *
   * @var \Drupal\media\MediaStorage
   */
  protected $fieldConfigStorage;

  /**
   * Role storage.
   *
   * @var \Drupal\user\RoleInterface
   */
  protected $roleStorage;

  /**
   * {@inheritdoc}
   *
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   * @throws \Drupal\Core\Entity\EntityStorageException
   */
  protected function setUp() {
    parent::setUp();
    $this->user = $this->drupalCreateUser([
      'administer pixabay settings',
      'administer entity browsers',
    ]);
    $this->drupalLogin($this->user);
    /** @var \Drupal\Core\Entity\EntityTypeManagerInterface $entityTypeManager */
    $entityTypeManager = $this->container->get('entity_type.manager');
    $this->mediaTypeStorage = $entityTypeManager->getStorage('media_type');
    $this->fieldConfigStorage = $entityTypeManager->getStorage('field_config');
    $this->roleStorage = $entityTypeManager->getStorage('user_role');
  }

  /**
   * Set up the media type with the image source plugin.
   */
  private function setUpMediaType() {
    $this->mediaType = $this->createMediaType('image', [
      'id' => 'test',
      'label' => 'Test',
    ]);
  }

  /**
   * Test Entity browser widget configuration and widget usage.
   *
   * @throws \Behat\Mink\Exception\ElementNotFoundException
   * @throws \Behat\Mink\Exception\ExpectationException
   * @throws \Behat\Mink\Exception\ResponseTextException
   */
  public function testPixabayWidget() {
    $webAssert = $this->assertSession();
    // Create a new entity browser...
    $this->drupalGet('/admin/config/content/entity_browser/add');
    $webAssert->fieldExists('label')->setValue('Test entity browser');
    $this->getSession()->executeScript("jQuery('.visually-hidden').removeClass('visually-hidden');");
    $webAssert->fieldExists('name')->setValue('test_entity_browser');
    $webAssert->buttonExists('Save')->press();
    $webAssert->addressEquals('/admin/config/content/entity_browser/test_entity_browser/widgets');
    $webAssert->pageTextContains('The entity browser Test entity browser has been added. Now you may configure the widgets you would like to use.');

    // ...check if the Pixabay plugin is available...
    $webAssert->responseContains('The available plugins are:');
    $webAssert->responseContains('<strong>Pixabay:</strong> Adds a Pixabay widget where images from Pixabay can be added as media to entities (use entity reference to media on entity fields). Only one Pixabay widget per entity browser is supported.');
    // ...add it...
    $this->addPixabayWidget($webAssert);
    // ...and check if correct responses are displayed (with no existing media
    // type with the image source field).
    $webAssert->responseContains("You don't have a media type with the Image source field.");

    // Setup a new media type with the image source
    // field and add the Pixabay widget again.
    $this->setUpMediaType();
    $this->drupalGet('/admin/config/content/entity_browser/test_entity_browser/widgets');
    $uuid = $this->addPixabayWidget($webAssert);

    // Check if all fields exist and have the correct default value.
    $webAssert->fieldValueEquals("table[$uuid][label]", 'pixabay');
    $webAssert->fieldValueEquals("table[$uuid][form][submit_text]", 'Select entities');
    $webAssert->fieldValueEquals("table[$uuid][form][search_placeholder]", 'Enter search terms...');
    $webAssert->fieldValueEquals("table[$uuid][form][search_text]", 'Pixabay search');
    $webAssert->fieldValueEquals("table[$uuid][form][upload_location]", 'public://Pixabay/[PIXABAY_SEARCH_TERM]/');
    $webAssert->responseContains('[PIXABAY_SEARCH_TERM] token will be replaced with the search term.');
    $webAssert->hiddenFieldValueEquals("table[$uuid][form][identifier]", 'pixabay-widget');
    $webAssert->fieldValueEquals("table[$uuid][form][extensions]", '');
    $webAssert->fieldValueEquals("table[$uuid][form][image_size]", '_640');

    // Test unsuccessful form submission (extensions validation).
    $uuid = $this->addPixabayWidget($webAssert);
    $this->selectMediaType($webAssert, $uuid);
    $this->drupalPostForm(NULL, ["table[$uuid][form][extensions]" => 'docx'], 'Save');
    $webAssert->responseContains('The allowed extensions on this widget do not match those that are allowed on the media type test, they might have been changed.
           Try saving the configuration again.');

    // Test deleting the widget.
    $webAssert->buttonExists("edit-table-$uuid-remove")->click();
    $webAssert->assertWaitOnAjaxRequest();
    $this->assertSession()->elementNotExists('css', 'tr[data-drupal-selector="edit-table-' . $uuid . '"]');

    // There should be no widgets now.
    $this->assertSession()->elementNotExists('css', 'tr.draggable');
    $this->drupalPostForm(NULL, [], 'Save');

    // Test the media type select ajax callback, which fills the extensions
    // field with the allowed extensions configured in the selected media type.
    // Test successful form submission.
    $uuid = $this->addPixabayWidget($webAssert);
    $this->selectMediaType($webAssert, $uuid);
    /** @var \Drupal\media\Entity\MediaType $mediaTypeLoaded */
    $mediaTypeLoaded = $this->mediaTypeStorage
      ->load('test');
    /** @var \Drupal\field\Entity\FieldConfig $fieldInstance */
    $fieldInstance = $this->fieldConfigStorage
      ->load('media.test.' . $mediaTypeLoaded->getSource()->getConfiguration()['source_field']);
    $extensions = $fieldInstance->get('settings')['file_extensions'];
    $webAssert->fieldValueEquals("table[$uuid][form][extensions]", $extensions);
    $this->drupalPostForm(NULL, [], 'Save');
    $webAssert->responseContains('The entity browser <em class="placeholder">Test entity browser</em> has been updated.');
    $webAssert->fieldExists("table[$uuid][form][extensions]")->setValue('jpg');
    $this->drupalPostForm(NULL, [], 'Save');
    $webAssert->responseContains('The entity browser <em class="placeholder">Test entity browser</em> has been updated.');
    $webAssert->fieldValueEquals("table[$uuid][form][extensions]", 'jpg');

    // Add the new permission for the test entity browser, that was created, to
    // our test user and navigate to the test entity browser form.
    $roleLoaded = $this->roleStorage->load(current($this->user->getRoles(TRUE)));
    $this->grantPermissions($roleLoaded, ['access test_entity_browser entity browser pages']);
    $this->drupalGet('/entity-browser/modal/test_entity_browser');

    // Test if the widget tab is active and has the correct name.
    $this->assertEqual($webAssert->elementExists('css', 'nav.eb-tabs a')
      ->getAttribute('class'), 'is-active active');
    $webAssert->elementContains('css', 'nav.eb-tabs a', 'pixabay');

    // Test if all elements exists with the correct attributes.
    $webAssert->elementExists('css', '#pixabay-wrapper');
    $pixabayLogoImage = $webAssert->elementExists('css', '.pixabay-logo-widget img');
    $this->assertEqual($pixabayLogoImage->getAttribute('alt'), 'Pixabay');
    $this->assertEqual($pixabayLogoImage->getAttribute('src'), 'https://pixabay.com/static/img/public/leaderboard_a.png');
    $pixabayLogoImageLink = $webAssert->elementExists('css', '.pixabay-logo-widget a');
    $this->assertEqual($pixabayLogoImageLink->getAttribute('href'), 'https://pixabay.com/');
    $this->assertEqual($pixabayLogoImageLink->getAttribute('target'), '_blank');
    $webAssert->responseContains('Pixabay search');
    $this->assertEqual($webAssert->fieldExists('search_key')->getAttribute('placeholder'), 'Enter search terms...');

    // Test empty search API key submit.
    $this->drupalPostForm(NULL, [], 'Pixabay search');
    $webAssert->responseContains('Please configure your Pixabay API search key.');

    // Test empty search term submit.
    $this->drupalPostForm('admin/config/media/pixabay', ['media_pixabay_api' => 'not-empty-string'], 'Save configuration');
    $this->drupalGet('/entity-browser/modal/test_entity_browser');
    $this->drupalPostForm(NULL, [], 'Pixabay search');
    $webAssert->responseContains('Please enter a search term.');

    // Test no results returned from api call.
    $this->config('media_pixabay_test.api')->set(
      'response',
      json_encode((object) $this->getMockResponseData('empty'))
    )->save();
    $this->drupalPostForm(NULL, ['search_key' => 'test'], 'Pixabay search');
    $webAssert->responseContains('Current search provided no results, please enter a valid search term.');

    // Test exception thrown in the api call.
    $this->config('media_pixabay_test.api')->set(
      'response',
      'exception'
    )->save();
    $this->drupalPostForm(NULL, ['search_key' => 'test'], 'Pixabay search');
    $webAssert->responseContains('The validation');
    $webAssert->responseContains('failed with error "test"');

    // Test valid data returned from the api call.
    $this->config('media_pixabay_test.api')->set(
      'response',
      json_encode((object) $this->getMockResponseData('valid'))
    )->save();
    $this->drupalPostForm(NULL, ['search_key' => 'test'], 'Pixabay search');

    // Test if classes are toggled correctly on click, element attributes and
    // extension validation.
    $webAssert->elementExists('css', '#result-images');
    $imageCheckbox = $this->selectImage($webAssert, '#edit-pictures-0', 'https://via.placeholder.com/1_640.png', 'https://via.placeholder.com/300.png', 150, 99);

    $this->drupalPostForm(NULL, [], 'Select entities');
    $this->assertEqual($imageCheckbox->hasClass('pixabay-configuration-error'), TRUE);
    $webAssert->responseContains('Extension png doesn\'t match the allowed extensions set in the widget configuration.');

    $imageCheckbox->click();
    $this->assertEqual($imageCheckbox->hasClass('pixabay-configuration-error'), FALSE);
    $this->selectImage($webAssert, '#edit-pictures-1', 'https://via.placeholder.com/1_640.jpg', 'https://via.placeholder.com/300.jpg', 160, 84);
    $this->drupalPostForm(NULL, [], 'Select entities');
    $webAssert->responseNotContains('Extension png doesn\'t match the allowed extensions set in the widget configuration.');

    // Test if results are correct with image size 1280.
    $this->drupalGet('/admin/config/content/entity_browser/test_entity_browser/widgets');
    $webAssert->fieldExists("table[$uuid][form][image_size]")->setValue('_1280');
    $this->drupalPostForm(NULL, [], 'Save');

    $this->drupalGet('/entity-browser/modal/test_entity_browser');
    $this->drupalPostForm(NULL, ['search_key' => 'test'], 'Pixabay search');
    $this->selectImage($webAssert, '#edit-pictures-0', 'https://via.placeholder.com/1_1280.png', 'https://via.placeholder.com/300.png', 150, 99);
  }

  /**
   * Select the media type.
   *
   * @param \Drupal\FunctionalJavascriptTests\WebDriverWebAssert $webAssert
   *   Web assert.
   * @param string $uuid
   *   Widget table uuid.
   *
   * @throws \Behat\Mink\Exception\ElementNotFoundException
   */
  private function selectMediaType(WebDriverWebAssert $webAssert, $uuid) {
    $mediaTypeSelect = $webAssert->selectExists("table[$uuid][form][media_type]");
    $mediaTypeSelect->selectOption('test');
    $webAssert->assertWaitOnAjaxRequest();
  }

  /**
   * Add the Pixabay widget.
   *
   * @param \Drupal\FunctionalJavascriptTests\WebDriverWebAssert $webAssert
   *   Web assert.
   *
   * @return string|string[]
   *   Returns the uuid of the added widget.
   *
   * @throws \Behat\Mink\Exception\ElementNotFoundException
   */
  private function addPixabayWidget(WebDriverWebAssert $webAssert) {
    $webAssert->selectExists('widget')->selectOption('pixabay');
    $selector = $webAssert
      ->waitforElementVisible('css', 'tr.draggable')
      ->getAttribute('data-drupal-selector');
    return str_replace('edit-table-', '', $selector);
  }

  /**
   * Select and test the image elements.
   *
   * @param \Drupal\FunctionalJavascriptTests\WebDriverWebAssert $webAssert
   *   Web assert.
   * @param string $selector
   *   Css selector.
   * @param string $imageURL
   *   Pixabay image url.
   * @param string $previewUrl
   *   Pixabay image preview url.
   * @param int $previewWidth
   *   Pixabay image preview width.
   * @param int $previewHeight
   *   Pixabay image preview height.
   *
   * @return \Behat\Mink\Element\NodeElement
   *   selected image checkbox element.
   *
   * @throws \Behat\Mink\Exception\ElementNotFoundException
   */
  public function selectImage(WebDriverWebAssert $webAssert, $selector, $imageURL, $previewUrl, $previewWidth, $previewHeight) {
    $imageCheckbox = $webAssert->elementExists('css', $selector);
    $imageCheckboxInput = $webAssert->elementExists('css', "$selector input");
    $image = $webAssert->elementExists('css', "$selector img");
    $this->assertEqual($image->hasClass('pixabay-image'), TRUE);
    $this->assertEqual($image->getAttribute('src'), $previewUrl);
    $this->assertEqual($image->getAttribute('alt'), 'Pixabay image');
    $this->assertEqual($image->getAttribute('width'), $previewWidth);
    $this->assertEqual($image->getAttribute('height'), $previewHeight);
    $this->assertEqual($imageCheckboxInput->isVisible(), FALSE);
    $this->assertEqual($imageCheckboxInput->getAttribute('value'), "$imageURL $previewUrl pixabay_id pixabay_tags");
    $this->assertEqual($imageCheckbox->isChecked(), FALSE);
    $imageCheckbox->click();
    $this->assertEqual($imageCheckboxInput->isChecked(), TRUE);
    $this->assertEqual($imageCheckbox->hasClass('pixabay-selected-image'), TRUE);
    return $imageCheckbox;
  }

  /**
   * Mocks returned data from the Pixabay API service.
   *
   * @param string $dataType
   *   Type of data to be returned.
   *
   * @return array
   *   Array with the mock data.
   */
  protected function getMockResponseData($dataType) {
    switch ($dataType) {
      case 'valid':
      default:
        return [
          "total" => 4692,
          "totalHits" => 500,
          "hits" => [
            [
              'id' => 'pixabay_id',
              'pageURL' => 'https://via.placeholder.com/300.png',
              'type' => 'photo',
              'tags' => 'pixabay_tags',
              'previewURL' => 'https://via.placeholder.com/300.png',
              'previewWidth' => 150,
              'previewHeight' => 99,
              'webformatURL' => 'https://via.placeholder.com/1_180.png',
              'webformatWidth' => 640,
              'webformatHeight' => 426,
              'largeImageURL' => 'https://via.placeholder.com/1_1280.png',
              'imageWidth' => 5184,
              'imageHeight' => 3456,
              'imageSize' => 2241942,
              'views' => 340044,
              'downloads' => 150358,
              'favorites' => 1021,
              'likes' => 646,
              'comments' => 141,
              'user_id' => 242387,
              'user' => 'Free-Photos',
              'userImageURL' => 'https://via.placeholder.com/300.png',
            ],
            [
              "id" => 'pixabay_id',
              "pageURL" => "https://via.placeholder.com/300.jpg",
              "type" => "photo",
              "tags" => "pixabay_tags",
              "previewURL" => "https://via.placeholder.com/300.jpg",
              "previewWidth" => 160,
              "previewHeight" => 84,
              "webformatURL" => "https://via.placeholder.com/1_180.jpg",
              "webformatWidth" => 640,
              "webformatHeight" => 360,
              "largeImageURL" => "https://via.placeholder.com/1_1280.jpg",
              "fullHDURL" => "https://via.placeholder.com/300.jpg",
              "imageURL" => "https://via.placeholder.com/300.jpg",
              "imageWidth" => 4000,
              "imageHeight" => 2250,
              "imageSize" => 4731420,
              "views" => 7671,
              "downloads" => 6439,
              "favorites" => 1,
              "likes" => 5,
              "comments" => 2,
              "user_id" => 48777,
              "user" => "Josch13",
              "userImageURL" => "https://via.placeholder.com/300",
            ],
          ],
        ];

      case 'empty':
        return ['hits' => NULL];
    }
  }

}
